from nolearn.lasagne import BatchIterator
import numpy as np
from scipy import misc
from PIL import Image

APS = 256;
PS = 227;

def rotate_img(x):
    rotate_angle = (np.random.rand(1)[0]-0.5)*2 * 20.0;
    im = Image.fromarray(x.astype(np.uint8));
    x = np.array(im.rotate(rotate_angle, Image.BICUBIC)).astype(np.float32);
    return x;

def data_aug_img(img, deterministic=False):
    # crop
    icut = APS - PS;
    jcut = APS - PS;
    if deterministic:
        ioff = int(icut // 2);
        joff = int(jcut // 2);
    else:
        ioff = np.random.randint(0, icut + 1);
        joff = np.random.randint(0, jcut + 1);
    img = img[:, ioff : ioff + img.shape[1] - icut, joff : joff + img.shape[2] - jcut];

    # adjust color
    if not deterministic:
        adj_range = 0.15;
        rgb_mean = np.mean(img, axis=(1,2), keepdims=True).astype(np.float32);
        adj_magn = np.random.uniform(1 - adj_range, 1 + adj_range, (3, 1, 1)).astype(np.float32);
        img = np.clip((img-rgb_mean)*adj_magn + rgb_mean + np.random.uniform(-1.0, 1.0, (3, 1, 1))*20, 0.0, 255.0);

    # mirror
    if not deterministic:
        if np.random.rand(1)[0] < 0.5:
            img = img[:, ::-1, :];

    # scaling
    if not deterministic:
        if np.random.rand(1)[0] < 0.5:
            iscale = 2*(np.random.rand(1)[0]-0.5)*0.10 + 1.0;
            jscale = 2*(np.random.rand(1)[0]-0.5)*0.10 + 1.0;
            img = misc.imresize(img.transpose().astype(np.uint8), \
                    (int(img.shape[2]*jscale), int(img.shape[1]*iscale)) \
                    ).transpose().astype(np.float32);

    # rotate small degree
    if not deterministic:
        if np.random.rand(1)[0] < 0.9:
            img = img.transpose();
            img = rotate_img(img);
            img = img.transpose();

    img = zero_centering(img);
    img = (img - 127.0) / 127.0;

    return img;

def zero_centering(img):
    x0 = (img.shape[1] - PS) // 2;
    y0 = (img.shape[2] - PS) // 2;
    im = Image.fromarray(img.transpose().astype(np.uint8));
    img = np.array(im.crop((x0, y0, x0+PS, y0+PS))).transpose().astype(np.float32);
    return img;

def data_aug(X, deterministic=False):
    Xc = np.zeros(shape=(X.shape[0], X.shape[1], PS, PS), dtype=np.float32);
    for i in range(len(X)):
        Xc[i] = data_aug_img(X[i], deterministic=deterministic);
    return Xc;

def inverse_transform(X):
    return (X * 127.0 + 127.0);

