import numpy as np
from lasagne import layers, objectives, nonlinearities
from nolearn.lasagne import NeuralNet
from nolearn.lasagne import BatchIterator
from nolearn.lasagne import TrainSplit
from constants import PS, fea_len, classn

def get_alex_net_small():
    cnn = NeuralNet(
        layers=[
            ('input', layers.InputLayer),

            ('conv1', layers.Conv2DLayer),
            ('pool1', layers.MaxPool2DLayer),
            ('norm1', layers.LocalResponseNormalization2DLayer),

            ('conv2', layers.Conv2DLayer),
            ('pool2', layers.MaxPool2DLayer),
            ('norm2', layers.LocalResponseNormalization2DLayer),

            ('conv3', layers.Conv2DLayer),
            ('pool3', layers.MaxPool2DLayer),

            ('fc6', layers.DenseLayer),
            ('drop_fc6', layers.DropoutLayer),
            ('fc7', layers.DenseLayer),
            ('drop_fc7', layers.DropoutLayer),
            ('output', layers.DenseLayer),
            ],
        input_shape=(None, 3, PS, PS),

        conv1_num_filters=96,
        conv1_filter_size=(7, 7),
        conv1_stride=4,
        pool1_pool_size=(3, 3),
        pool1_stride=(2, 2),

        conv2_num_filters=256,
        conv2_filter_size=(3, 3),
        conv2_pad=(2, 2),
        pool2_pool_size=(3, 3),
        pool2_stride=(2, 2),

        conv3_num_filters=384,
        conv3_filter_size=(3, 3),
        conv3_pad=(1, 1),
        pool3_pool_size=(3, 3),
        pool3_stride=(2, 2),

        fc6_num_units=512,
        drop_fc6_p=0.5,
        fc7_num_units=fea_len,
        drop_fc7_p=0.5,

        output_num_units=classn,
        output_nonlinearity=nonlinearities.softmax,

        update_learning_rate=1e-10,
        update_momentum=0.99,
        objective_loss_function=objectives.squared_error,
        batch_iterator_train=BatchIterator(batch_size=1),
        batch_iterator_test=BatchIterator(batch_size=1),
        train_split=TrainSplit(eval_size=0.5),
        regression=True,
        max_epochs=1,
        verbose=1,
    );

    # initialize the nolearn network
    X = np.zeros((2, 3, PS, PS), dtype=np.float32);
    y = np.zeros((2, classn), dtype=np.float32);
    y[0, 0] = 1;
    y[1, 0] = 1;
    cnn.fit(X, y);

    return cnn;

